#include "world.h"
	

Sphere::Sphere()
{
}

Sphere::Sphere(Vector3 v) : GLObject(v)
{
}

void Sphere::Render()
{
	save();

	transform();
	light();

	GLObject::Render();

	GLUquadric* quadratic = gluNewQuadric();
	gluSphere(quadratic, 1, 20, 20);
	gluDeleteQuadric(quadratic);

	restore();
}

Light::Light() : lightId(0)
{
	setAmbient(0.3f, 0.3f, 0.3f);
	setScale(2, 2, 2);	// Makes lights a bit larger.

	Linear() = 0.02f;
}

Light::Light(Vector3 v) : Sphere(v), lightId(0)
{
	setAmbient(0.3f, 0.3f, 0.3f);
	setScale(2, 2, 2);	// Makes lights a bit larger.

	Linear() = 0.02f;
}

Light::Light(int id) :  lightId(id)
{
	setAmbient(0.3f, 0.3f, 0.3f);
	setScale(2, 2, 2);	// Makes lights a bit larger.

	Linear() = 0.02f;
}

int& Light::id()
{
	return lightId;
}

float& Light::Linear()
{
	return linear;
}

void Light::setAmbient(float r, float g, float b, float a)
{
	ambient[0] = r;
	ambient[1] = g;
	ambient[2] = b;
	ambient[3] = a;
}

void Light::Render()
{
	float pos[] = {m_X, m_Y, m_Z, 1.0f};

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0 + lightId, GL_POSITION, pos);
	glLightfv(GL_LIGHT0 + lightId, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0 + lightId, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT0 + lightId, GL_SPECULAR, specular);
	glLightf(GL_LIGHT0 + lightId, GL_LINEAR_ATTENUATION, linear);

	glEnable(GL_LIGHT0 + lightId);

	//

	save();

	GLObject::Render();

	transform();

	glDisable(GL_LIGHTING);

	glColor3f(1.0f, 0.5f, 0.0f);

	// Turn below on for blinking light source

                                            //speed  //min brightness  //max brightness          //area of brightness
	/*float c = (cosf((float)GetTickCount() / 500.0f) + 2.5f)            * 0.25f              +   0.25f;
		
		glColor3f(c, c, 0.0f);

		setDiffuse(c, c, c);*/

	GLUquadric* quadratic = gluNewQuadric();
	gluSphere(quadratic, 1, 20, 20);
	gluDeleteQuadric(quadratic);
	
	restore();
}

ControlPoint::ControlPoint() : selected(false)
{
}

ControlPoint::ControlPoint(Vector3 v) : Sphere(v), selected(false)
{
}

void ControlPoint::Render()
{	
	save();
	
	if(selected)
	{
		float t = (sinf((float)GetTickCount() / 250.0f) + 1.0f)*0.25f + 0.5f;
		
		glColor3f(0.0f, t, 0.0f);

		setDiffuse(0.0f, t, 0.0f);
		setScale(t/4.0f + 0.5f);
	}
	else
	{
		setScale(0.5f);
		setDiffuse(0.0f, 0.0f, 1.0f);
	}
	
	Sphere::Render();

	restore();
}

void ControlPoint::select()
{
	selected = true;
}

void ControlPoint::deSelect()
{
	selected = false;
}

Fighter::Fighter() : At(0, 1, 0), Up(0, 0, 1), trackingCamera(0)
{
}

Fighter::Fighter(Vector3 v) : Mesh(v), At(0, 1, 0), Up(0, 0, 1), trackingCamera(0)
{
}

void Fighter::Destroy()
{
	Mesh::Destroy();
}

void Fighter::Create()
{
	LoadOFF("Data/Fighter.off");
	LoadTexture("Data/Fighter.bmp");

	setScale(0.2f, 0.2f, 0.2f);

	trackingCamera = NULL;
}

void Fighter::Render()
{
	float mat[16];
	float transposed[16];

	save();

	// Here the camera transformation matrix is used, with a centered eye.
	// Because the look-at matrix is the inverse of world rotation, we takes its inverse.
	// The inverve of a matrix with only rotation is its transpose.

	// This assumes the model is facing the z-direction with a y up vector.

	Vector3 TransAt = At - (*this);	// We are calculating from the center.

	glPushMatrix();

	glLoadIdentity();
	gluLookAt(0, 0, 0, TransAt.m_X, TransAt.m_Y, TransAt.m_Z, Up.m_X, Up.m_Y, Up.m_Z);
	glGetFloatv(GL_MODELVIEW_MATRIX, mat);

	// Transpose.

	for(int i = 0; i < 4; i++)
		for(int j = 0; j < 4; j++)
			transposed[i*4+j] = mat[j*4+i];

	glPopMatrix();

	// Finaly move it into place.
	
	transform();
	light();

	// Then align with the inverse look at matrix.

	glMultMatrixf(transposed);

	// Align the fighter to face in the (0, 1, 0) direction.

	glRotatef(-90.0f, 0.0f, 0.0f, 1.0f);
	glRotatef(-90.0f, 0.0f, 1.0f, 0.0f);

	Mesh::Render();

	restore();
}

Vector3& Fighter::at()
{
	return At;
}

void Fighter::Track(Camera* camera)
{
	this->trackingCamera = camera;
}

void Fighter::AlignToCamera()
{
	if(!trackingCamera) return;
	
	trackingCamera->Position()	= *(Vector3*)this + Vector3(0.0f, 0.0f, 2.0f);	// Camera just above the fighter.
	trackingCamera->Forward()	= At - *(Vector3*)this;
}

// Assign:
// Light id's and path w.
// Time step and velocity.
// Gravity as on Earth.
// We are not paused.
// Physics on.
// Not looped.
World::World() : light1(0), light2(1), light3(2),
	w(0), dt(0.1f), u(0.0f), g(-9.8f),
	paused(false), physicsEnabled(true), looped(false)
{
	srand(time(0)); //randomise stuff
}

void World::Destroy()
{
	spline.clear(); // Get rid of all the spline points.

	// Instruct all objects to release the data they hold.
	for(int i = 0; i < (int)size(); i++) at(i)->Destroy();

	// Now clear the actaul object list.
	clear(); // Delete all our objects.

	paused			= false;
	physicsEnabled	= true;
	looped			= false;
}

void World::Create(Camera* pCamera)
{
	// First clean up...

	Destroy();

	// Now (re)create everything...

	spline = BezierSpline();

	for(int i = -5; i < 5; i++)
	{
		//Vector3 v(i*5.0f, 0, 0);	// Space out the points nicely.
		Vector3 v(i*10, -(i*rand()%25), -(i*rand()%25));	// Generate random spline to start with
		spline.push_back(new ControlPoint(v));
	}

	(Vector3&)spline = Vector3(0, 0, +50);	// Move the spline up a bit.

	//

	fighter.Create();

	//

	//lights position
	light1.setDiffuse(1.0f, 1.0f, 0.9f);
	light1.m_Z = 60;
	light1.m_X = -35;
	light1.m_Y = 15;

	light2.setDiffuse(1.0f, 1.0f, 0.8f);
	light2.m_X = 15;
	light2.m_Y = 30;
	light2.m_Z = 25;

	light3.setDiffuse(1.0f, 0.5f, 0.6f);
	light3.m_X = 4;
	light3.m_Y = 20;
	light3.m_Z = 65;

	//

	glEnable(GL_LIGHTING);	// We want lighting.

	terrain.Load("Data/Terrain.ppm", "Data/Terrain.bmp");
	skybox.Load(pCamera);

	AlignFighter();
	
	push_back(&skybox);		// Skybox first for z-buffer purposes.
	push_back(&light1);
	push_back(&light2);
	push_back(&light3);
	push_back(&terrain);
	push_back(&fighter);
	push_back(&spline);		// Spline last because of visible control points.
	
	// Animation...

	dt	= 0.1f;
	u	= 0.0f;
	w	= 0.0f;
	g	= -9.8f;		// Earth gravity.
}

void World::Render()
{		
	// Render all our objects.
	for(int i = 0; i < (int)size(); i++) at(i)->Render();
}

/*

This is the main picking method. The basic idea is to render the scene using a pick camera matrix.
The matrix constrains the rendering to a small area around the cursor. The scene is then rendered as normal,
except for that each object writes a unique ID (a pointer to itself in this case).

A list is then returned with all the objects which intersected the cursor ray. The nearest one to the screen
is found and cast to a GLObject and returned.

*/

GLObject* World::PickObject(Camera& camera, int x, int y)	// Use the camera to pick
															// out an object based on a viewport point.
{
	const int BUFSIZE = 128;	// Max 128 objects
	
	GLuint selectBuffer[BUFSIZE];
	GLint hits;

	glSelectBuffer(BUFSIZE, selectBuffer);
	
	glRenderMode(GL_SELECT);

	// Render the scene with the pick matrix, confining rendering to the cursor.
	
	camera.render(x, y);

	glInitNames();
	glPushName(0);
	
	// Render the scene so we can get a list of picked objects.

	Render();

	hits = glRenderMode(GL_RENDER);	// Switch back to standard rendering
									// This also returns the number of hits with the last GL_SELECT render.

	cout << "Hits: " << hits << "\n";

	GLuint* ptr;
	ptr = (GLuint *) selectBuffer;

	GLuint nearestZ		= 0xffffffff;
	GLuint nearestId	= 0;

	for (int i = 0; i < hits; i++) 
	{
		// Record contains:
		//
		// Number of names for this hit
		// Near z
		// Far z
		// Names for hit

		// We are only interested in the nearest id.
		
		GLuint names	= ptr[0];
		GLuint nearZ	= ptr[1];
		GLuint farZ		= ptr[2];

		cout << "Hit " << i+1 << endl;
		cout << "  Names\t\t= " << names << endl;
		cout << "  Near Z\t= " << nearZ << endl;
		cout << "  Far Z\t\t= " << farZ << endl;
		
		if(nearZ < nearestZ)
		{
			nearestZ = ptr[1];
			nearestId = ptr[3+0];	// First id.
		}

		ptr += (3 + names);	// Skip the name records.
	}

	return nearestId == 0 ? 0 : (GLObject*)nearestId;
}

bool World::IsControlPoint(GLObject* obj)
{
	vector<GLObject*>::iterator it = find(spline.begin(), spline.end(), obj);

	return (it != spline.end());
}

bool World::IsLight(GLObject* obj)
{
	return (obj == &light1 || obj == &light2 || obj == &light3);
}


/*

This is where the physics is done. 

*/

void World::Step()
{		
	AlignFighter();	// Even if we are paused, we should align the fighter.

	if(Paused()) return;

	// We need to move a set distance, but because the spline is more densely populated in different
	// areas, we have to normalise that first.

	float s = 0;		// How far travelled
	float ds = 0;		// How far left
	
	// Calculate gravity acceleration vector.
	float a = (spline.f(w+0.05f) - spline.f(w)).Normalized().Dot(Vector3(0.0f, 0.0f, g));

	if(PhysicsEnabled()) u += 0.5f*a*dt;
	ds = u*dt;

	if(PhysicsEnabled()) u *= 0.99f; // Friction.

	if(!PhysicsEnabled()) u = 5.0f;	// Some default velocity.

	// This filters off "shaky" movement by not moving is the distance is too little.

	if(abs(ds) > 0.0005f) while(abs(s) < abs(ds))
	{
		Vector3 current = spline.f(w);

		w += ds * 0.0001f;	// The amount to step on the curve.

		s += (spline.f(w) - current).Length();

		if(w > 1.0f) { w = 1.0f; u=0.0f; }
		if(w < 0.0f) { w = 0.0f; u=0.0f; }
	}

	if(Looped() && w >= 1.0f)
	{
		u = 5.0f;	// Give it some initial velocity.
		w = 0.0f;
	}
}

void World::AlignFighter()	// Align the nose of the fighter correctly.
{
	// Place the model on the path.
	(Vector3&)fighter	= spline + spline.f(w);

	// Forward is slightly ahead of current.
	fighter.at() =  spline + spline.f(w+0.02f);
}

bool& World::Paused()
{
	return paused;
}

bool& World::PhysicsEnabled()
{
	return physicsEnabled;
}

bool& World::Looped()
{
	return looped;
}